# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_schemas.base import BaseOneOfSchema
from polyaxon_schemas.ops.build_job import BuildConfig, BuildSchema
from polyaxon_schemas.ops.experiment import ExperimentConfig, ExperimentSchema
from polyaxon_schemas.ops.group import GroupConfig, GroupSchema
from polyaxon_schemas.ops.job import JobConfig, JobSchema
from polyaxon_schemas.ops.notebook import NotebookConfig, NotebookSchema
from polyaxon_schemas.ops.tensorboard import TensorboardConfig, TensorboardSchema


class TemplateSchema(BaseOneOfSchema):
    TYPE_FIELD = 'kind'
    TYPE_FIELD_REMOVE = False
    SCHEMAS = {
        BuildConfig.IDENTIFIER: BuildSchema,
        JobConfig.IDENTIFIER: JobSchema,
        ExperimentConfig.IDENTIFIER: ExperimentSchema,
        GroupConfig.IDENTIFIER: GroupSchema,
        TensorboardConfig.IDENTIFIER: TensorboardSchema,
        NotebookConfig.IDENTIFIER: NotebookSchema,
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from collections import Mapping

from hestia.cached_property import cached_property
from marshmallow import EXCLUDE

from polyaxon_schemas.exceptions import PolyaxonConfigurationError
from polyaxon_schemas.ops.job import JobConfig
from polyaxon_schemas.ops.run import RunConfig
from polyaxon_schemas.specs import kinds
from polyaxon_schemas.specs.base import BaseRunSpecification, BaseSpecification


class JobSpecification(BaseRunSpecification):
    """The polyaxonfile specification for run jobs.

    SECTIONS:
        VERSION: defines the version of the file to be parsed and validated.
        LOGGING: defines the logging
        TAGS: defines the tags
        ENVIRONMENT: defines the run environment for experiment.
        BUILD: defines the build step where the user can set a docker image definition
        RUN: defines the run step where the user can run a command
    """
    _SPEC_KIND = kinds.JOB

    REQUIRED_SECTIONS = BaseRunSpecification.REQUIRED_SECTIONS + (
        BaseSpecification.BUILD, BaseSpecification.RUN,
    )

    POSSIBLE_SECTIONS = BaseRunSpecification.POSSIBLE_SECTIONS + (
        BaseRunSpecification.PARAMS,
        BaseRunSpecification.RUN,
    )

    CONFIG = JobConfig

    @cached_property
    def run(self):
        return self.config.run

    @classmethod
    def create_specification(cls,  # pylint:disable=arguments-differ
                             build_config,
                             run_config,
                             to_dict=True):
        try:
            specification = BaseRunSpecification.create_specification(
                build_config=build_config, to_dict=True)
        except PolyaxonConfigurationError:
            raise PolyaxonConfigurationError(
                'Create specification expects a dict or an instance of BuildConfig.')

        if isinstance(run_config, RunConfig):
            r_config = run_config.to_light_dict()
        elif isinstance(run_config, Mapping):
            r_config = RunConfig.from_dict(run_config, unknown=EXCLUDE)
            r_config = r_config.to_light_dict()
        else:
            raise PolyaxonConfigurationError(
                'Create specification expects a dict or an instance of RunConfig.')

        specification[cls.KIND] = cls._SPEC_KIND
        specification[cls.RUN] = r_config

        if to_dict:
            return specification
        return cls.read(specification)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import copy

from polyaxon_schemas.exceptions import PolyaxonfileError
from polyaxon_schemas.ops.build_job import BuildConfig
from polyaxon_schemas.ops.group.hptuning import HPTuningConfig
from polyaxon_schemas.ops.logging import LoggingConfig
from polyaxon_schemas.ops.run import RunConfig


def validate_headers(spec, data):
    """Validates headers data and creates the config objects"""
    validated_data = {
        spec.VERSION: data[spec.VERSION],
        spec.KIND: data[spec.KIND],
    }

    if data.get(spec.LOGGING):
        validated_data[spec.LOGGING] = LoggingConfig.from_dict(
            data[spec.LOGGING])

    if data.get(spec.TAGS):
        validated_data[spec.TAGS] = data[spec.TAGS]

    if data.get(spec.HP_TUNING):
        validated_data[spec.HP_TUNING] = HPTuningConfig.from_dict(
            data[spec.HP_TUNING])

    return validated_data


def validate(spec, data):
    """Validates the data and creates the config objects"""
    data = copy.deepcopy(data)
    validated_data = {}

    def validate_keys(section, config, section_data):
        extra_args = [key for key in section_data.keys() if key not in config.SCHEMA().fields]
        if extra_args:
            raise PolyaxonfileError('Extra arguments passed for `{}`: {}'.format(
                section, extra_args))

    def add_validated_section(section, config):
        if data.get(section):
            section_data = data[section]
            validate_keys(section=section, config=config, section_data=section_data)
            validated_data[section] = config.from_dict(section_data)

    add_validated_section(spec.ENVIRONMENT, spec.ENVIRONMENT_CONFIG)
    add_validated_section(spec.BUILD, BuildConfig)
    add_validated_section(spec.RUN, RunConfig)

    return validated_data



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_layers

from polyaxon_schemas.ml.initializations import (
    GlorotNormalInitializerConfig,
    OrthogonalInitializerConfig,
    ZerosInitializerConfig
)
from polyaxon_schemas.ml.layers.convolutional_recurrent import (
    ConvLSTM2DConfig,
    ConvRecurrent2DConfig
)


class TestConvolutionalRecurrentConfigs(TestCase):
    def test_conv_recurrent_2d_config(self):
        config_dict = {
            'filters': 20,
            'kernel_size': 3,
            'strides': [1, 1],
            'padding': 'valid',
            'data_format': None,
            'dilation_rate': [1, 1],
            'return_sequences': False,
            'go_backwards': False,
            'stateful': False
        }
        config = ConvRecurrent2DConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_conv_lstm_2d_config(self):
        config_dict = {
            'filters': 20,
            'kernel_size': 3,
            'strides': [1, 1],
            'padding': 'valid',
            'data_format': None,
            'dilation_rate': [1, 1],
            'activation': 'tanh',
            'recurrent_activation': 'hard_sigmoid',
            'use_bias': True,
            'kernel_initializer': GlorotNormalInitializerConfig().to_schema(),
            'recurrent_initializer': OrthogonalInitializerConfig().to_schema(),
            'bias_initializer': ZerosInitializerConfig().to_schema(),
            'unit_forget_bias': True,
            'kernel_regularizer': None,
            'recurrent_regularizer': None,
            'bias_regularizer': None,
            'activity_regularizer': None,
            'kernel_constraint': None,
            'recurrent_constraint': None,
            'bias_constraint': None,
            'return_sequences': False,
            'go_backwards': False,
            'stateful': False,
            'dropout': 0.,
            'recurrent_dropout': 0.
        }
        config = ConvLSTM2DConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_dict, assert_tensors

from polyaxon_schemas.ml.losses import (
    AbsoluteDifferenceConfig,
    ClippedDeltaLossConfig,
    CosineDistanceConfig,
    HingeLossConfig,
    HuberLossConfig,
    KullbackLeiberDivergenceConfig,
    LogLossConfig,
    MeanSquaredErrorConfig,
    PoissonLossConfig,
    SigmoidCrossEntropyConfig,
    SoftmaxCrossEntropyConfig
)


class TestLossConfigs(TestCase):
    @staticmethod
    def assert_equal_losses(l1, l2):
        assert_tensors(l1.pop('input_layer', None), l2.pop('input_layer', None))
        assert_tensors(l1.pop('output_layer', None), l2.pop('output_layer', None))
        assert_equal_dict(l1, l2)

    def test_base_losses_config(self):
        config_dict = {
            'input_layer': 'images',
            'output_layer': 'relu_1',
            'weights': 1.0,
            'name': 'l',
            'collect': False
        }

        config_classes = [
            AbsoluteDifferenceConfig,
            MeanSquaredErrorConfig,
            HingeLossConfig
        ]

        for config_class in config_classes:
            config = config_class.from_dict(config_dict)
            self.assert_equal_losses(config.to_dict(), config_dict)

    def test_log_loss_config(self):
        config_dict = {
            'input_layer': 'images',
            'output_layer': 'relu_1',
            'epsilon': 0.0001,
            'weights': 1.0,
            'name': 'l',
            'collect': False
        }
        config = LogLossConfig.from_dict(config_dict)
        self.assert_equal_losses(config.to_dict(), config_dict)

    def test_clipped_loss_config(self):
        config_dict = {
            'input_layer': 'images',
            'output_layer': 'relu_1',
            'clip_value_min': -0.1,
            'clip_value_max': -0.1,
            'weights': 1.0,
            'name': 'l',
            'collect': False
        }
        config = ClippedDeltaLossConfig.from_dict(config_dict)
        self.assert_equal_losses(config.to_dict(), config_dict)

    def test_huber_loss_config(self):
        config_dict = {
            'input_layer': 'images',
            'output_layer': 'relu_1',
            'clip': 0.1,
            'weights': 1.0,
            'name': 'l',
            'collect': False
        }
        config = HuberLossConfig.from_dict(config_dict)
        self.assert_equal_losses(config.to_dict(), config_dict)

    def test_softmax_crossentropy_loss_config(self):
        config_dict = {
            'input_layer': 'images',
            'output_layer': 'relu_1',
            'label_smoothing': 0.,
            'weights': 1.0,
            'name': 'l',
            'collect': False
        }
        config = SoftmaxCrossEntropyConfig.from_dict(config_dict)
        self.assert_equal_losses(config.to_dict(), config_dict)

    def test_sigmoid_crossentropy_loss_config(self):
        config_dict = {
            'input_layer': 'images',
            'output_layer': 'relu_1',
            'label_smoothing': 0.,
            'weights': 1.0,
            'name': 'l',
            'collect': False
        }
        config = SigmoidCrossEntropyConfig.from_dict(config_dict)
        self.assert_equal_losses(config.to_dict(), config_dict)

    def test_cosine_distance_loss_config(self):
        config_dict = {
            'input_layer': 'images',
            'output_layer': 'relu_1',
            'dim': 0,
            'weights': 1.0,
            'name': 'l',
            'collect': False
        }
        config = CosineDistanceConfig.from_dict(config_dict)
        self.assert_equal_losses(config.to_dict(), config_dict)

    def test_poisson_loss_config(self):
        config_dict = {
            'input_layer': 'images',
            'output_layer': 'relu_1',
            'weights': 1.0,
            'name': 'l',
            'collect': False
        }
        config = PoissonLossConfig.from_dict(config_dict)
        self.assert_equal_losses(config.to_dict(), config_dict)

    def test_kullbackleiber_div_loss_config(self):
        config_dict = {
            'input_layer': 'images',
            'output_layer': 'relu_1',
            'dim': 0,
            'weights': 1.0,
            'name': 'l',
            'collect': False
        }
        config = KullbackLeiberDivergenceConfig.from_dict(config_dict)
        self.assert_equal_losses(config.to_dict(), config_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from marshmallow import Schema, ValidationError, fields

from polyaxon_schemas.base import BaseConfig, BaseOneOfSchema, BaseSchema

REQUIRED_ERROR = u'Missing data for required field.'


class FooSchema(BaseSchema):
    value = fields.String(required=True)

    @staticmethod
    def schema_config():
        return FooConfig


class FooConfig(BaseConfig):
    SCHEMA = FooSchema
    IDENTIFIER = 'foo'

    def __init__(self, value=None):
        self.value = value

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.value == other.value


class BarSchema(BaseSchema):
    value = fields.Integer(required=True)

    @staticmethod
    def schema_config():
        return BarConfig


class BarConfig(BaseConfig):
    SCHEMA = BarSchema
    IDENTIFIER = 'bar'

    def __init__(self, value=None):
        self.value = value

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.value == other.value


class BazSchema(BaseSchema):
    value1 = fields.Integer(required=True)
    value2 = fields.String(required=True)

    @staticmethod
    def schema_config():
        return BazConfig


class BazConfig(BaseConfig):
    SCHEMA = BazSchema
    IDENTIFIER = 'baz'

    def __init__(self, value1=None, value2=None):
        self.value1 = value1
        self.value2 = value2

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.value1 == other.value1 and
                self.value2 == other.value2)


class EmptySchema(BaseSchema):

    @staticmethod
    def schema_config():
        return EmptyConfig


class EmptyConfig(BaseConfig):
    SCHEMA = EmptySchema
    IDENTIFIER = 'empty'


class MySchema(BaseOneOfSchema):
    SCHEMAS = {
        'foo': FooSchema,
        'bar': BarSchema,
        'baz': BazSchema,
        'empty': EmptySchema,
    }


class TestOneOfSchema(TestCase):
    def test_dump(self):
        foo_result = MySchema().dump(FooConfig('hello'))
        assert {'type': 'foo', 'value': 'hello'} == foo_result

        bar_result = MySchema().dump(BarConfig(123))
        assert {'type': 'bar', 'value': 123} == bar_result

    def test_dump_many(self):
        result = MySchema().dump([FooConfig('hello'), BarConfig(123)], many=True)
        assert [{'type': 'foo', 'value': 'hello'},
                {'type': 'bar', 'value': 123}] == result

    def test_dump_many_in_constructor(self):
        result = MySchema(many=True).dump([FooConfig('hello'), BarConfig(123)])
        assert [{'type': 'foo', 'value': 'hello'},
                {'type': 'bar', 'value': 123}] == result

    def test_dump_with_empty_keeps_type(self):
        result = MySchema().dump(EmptyConfig())
        assert {'type': 'empty'} == result

    def test_load(self):
        foo_result = MySchema().load({'type': 'foo', 'value': 'world'})
        assert FooConfig('world') == foo_result

        bar_result = MySchema().load({'type': 'bar', 'value': 456})
        assert BarConfig(456) == bar_result

    def test_load_many(self):
        result = MySchema().load([
            {'type': 'foo', 'value': 'hello world!'},
            {'type': 'bar', 'value': 123},
        ], many=True)
        assert FooConfig('hello world!'), BarConfig(123) == result

    def test_load_many_in_constructor(self):
        result = MySchema(many=True).load([
            {'type': 'foo', 'value': 'hello world!'},
            {'type': 'bar', 'value': 123},
        ])
        assert FooConfig('hello world!'), BarConfig(123) == result

    def test_load_removes_type_field(self):
        class Nonlocal(object):
            data = None

        class MySchema(Schema):
            def load(self, data, *args, **kwargs):
                Nonlocal.data = data
                return super(MySchema, self).load(data, *args, **kwargs)

        class FooSchema(MySchema):
            foo = fields.String(required=True)

        class BarSchema(MySchema):
            bar = fields.Integer(required=True)

        class TestSchema(BaseOneOfSchema):
            SCHEMAS = {
                'foo': FooSchema,
                'bar': BarSchema,
            }

        TestSchema().load({'type': 'foo', 'foo': 'hello'})
        assert 'type' not in Nonlocal.data

        TestSchema().load({'type': 'bar', 'bar': 123})
        assert 'type' not in Nonlocal.data

    def test_load_keeps_type_field(self):
        class Nonlocal(object):
            data = None
            type = None

        class MySchema(Schema):
            def load(self, data, *args, **kwargs):
                Nonlocal.data = data
                return super(MySchema, self).load(data, *args, **kwargs)

        class FooSchema(MySchema):
            foo = fields.String(required=True)

        class BarSchema(MySchema):
            bar = fields.Integer(required=True)

        class TestSchema(BaseOneOfSchema):
            TYPE_FIELD_REMOVE = False
            SCHEMAS = {
                'foo': FooSchema,
                'bar': BarSchema,
            }

        TestSchema(unknown='exclude').load({'type': 'foo', 'foo': 'hello'})
        assert Nonlocal.data['type'] == 'foo'

        TestSchema(unknown='exclude').load({'type': 'bar', 'bar': 123})
        assert Nonlocal.data['type'] == 'bar'

    def test_load_non_dict(self):
        with self.assertRaises(ValidationError):
            MySchema().load(123)

        with self.assertRaises(ValidationError):
            MySchema().load('foo')

    def test_load_errors_no_type(self):
        with self.assertRaises(ValidationError):
            MySchema().load({'value': 'foo'})

    def test_load_errors_field_error(self):
        with self.assertRaises(ValidationError):
            MySchema().load({'type': 'foo'})

    def test_load_errors_strict(self):
        with self.assertRaises(ValidationError):
            MySchema().load({'type': 'foo'})

    def test_load_many_errors_are_indexed_by_object_position(self):
        with self.assertRaises(ValidationError):
            MySchema().load([
                {'type': 'foo'},
                {'type': 'bar', 'value': 123},
            ], many=True)

    def test_load_many_errors_strict(self):
        with self.assertRaises(ValidationError):
            MySchema().load([
                {'type': 'foo', 'value': 'hello world!'},
                {'type': 'foo'},
                {'type': 'bar', 'value': 123},
                {'type': 'bar', 'value': 'hello'},
            ], many=True)

    def test_load_partial_specific(self):
        result = MySchema().load(
            {'type': 'foo'},
            partial=('value', 'value2',)
        )
        assert FooConfig() == result

        result = MySchema().load(
            {'type': 'baz', 'value1': 123},
            partial=('value', 'value2',)
        )
        assert BazConfig(value1=123) == result

    def test_load_partial_any(self):
        result = MySchema().load({'type': 'foo'}, partial=True)
        assert FooConfig() == result

        result = MySchema().load({'type': 'baz', 'value1': 123}, partial=True)
        assert BazConfig(value1=123) == result

        result = MySchema().load({'type': 'baz', 'value2': 'hello'}, partial=True)
        assert BazConfig(value2='hello') == result

    def test_load_partial_specific_in_constructor(self):
        result = MySchema(partial=('value', 'value2',)).load({'type': 'foo'})
        assert FooConfig() == result

        result = MySchema(partial=('value', 'value2',)) \
            .load({'type': 'baz', 'value1': 123})
        assert BazConfig(value1=123) == result

    def test_load_partial_any_in_constructor(self):
        result = MySchema(partial=True).load({'type': 'foo'})
        assert FooConfig() == result

        result = MySchema(partial=True).load({'type': 'baz', 'value1': 123})
        assert BazConfig(value1=123) == result

        result = MySchema(partial=True).load({'type': 'baz', 'value2': 'hello'})
        assert BazConfig(value2='hello') == result

    def test_validate(self):
        assert {} == MySchema().validate({'type': 'foo', 'value': '123'})
        assert {
                   0: {'value': [REQUIRED_ERROR]}
               } == MySchema().validate({'type': 'bar'})
        assert {
                   0: {'value': [REQUIRED_ERROR]}
               } == MySchema().validate({'type': 'bar'})

    def test_validate_many(self):
        errors = MySchema().validate([
            {'type': 'foo', 'value': '123'},
            {'type': 'bar', 'value': 123},
        ], many=True)
        assert {} == errors

        errors = MySchema().validate([
            {'value': '123'},
            {'type': 'bar'},
        ], many=True)
        assert {0: {'type': [REQUIRED_ERROR]},
                1: {'value': [REQUIRED_ERROR]}} == errors

        errors = MySchema().validate([
            {'value': '123'},
            {'type': 'bar'},
        ], many=True)
        assert {0: {'type': [REQUIRED_ERROR]},
                1: {'value': [REQUIRED_ERROR]}} == errors

    def test_validate_many_in_constructor(self):
        errors = MySchema(many=True).validate([
            {'type': 'foo', 'value': '123'},
            {'type': 'bar', 'value': 123},
        ])
        assert {} == errors

        errors = MySchema(many=True).validate([
            {'value': '123'},
            {'type': 'bar'},
        ])
        assert {0: {'type': [REQUIRED_ERROR]},
                1: {'value': [REQUIRED_ERROR]}} == errors

    def test_validate_partial_specific(self):
        errors = MySchema().validate(
            {'type': 'foo'},
            partial=('value', 'value2',)
        )
        assert {} == errors

        errors = MySchema().validate(
            {'type': 'baz', 'value1': 123},
            partial=('value', 'value2',)
        )
        assert {} == errors

    def test_validate_partial_any(self):
        errors = MySchema().validate(
            {'type': 'foo'}, partial=True
        )
        assert {} == errors

        errors = MySchema().validate(
            {'type': 'baz', 'value1': 123},
            partial=True
        )
        assert {} == errors

        errors = MySchema().validate(
            {'type': 'baz', 'value2': 'hello'},
            partial=True
        )
        assert {} == errors

    def test_validate_partial_specific_in_constructor(self):
        errors = MySchema(
            partial=('value', 'value2',)
        ).validate({'type': 'foo'})
        assert {} == errors

        errors = MySchema(
            partial=('value', 'value2',)
        ).validate({'type': 'baz', 'value1': 123})
        assert {} == errors

    def test_validate_partial_any_in_constructor(self):
        errors = MySchema(
            partial=True
        ).validate({'type': 'foo'})
        assert {} == errors

        errors = MySchema(
            partial=True
        ).validate({'type': 'baz', 'value1': 123})
        assert {} == errors

        errors = MySchema(
            partial=True
        ).validate({'type': 'baz', 'value2': 'hello'})
        assert {} == errors

    def test_using_as_nested_schema(self):
        class SchemaWithList(Schema):
            items = fields.List(fields.Nested(MySchema))

        schema = SchemaWithList()
        result = schema.load({'items': [
            {'type': 'foo', 'value': 'hello world!'},
            {'type': 'bar', 'value': 123},
        ]})
        assert {'items': [FooConfig('hello world!'), BarConfig(123)]} == result

        with self.assertRaises(ValidationError):
            schema.load({'items': [
                {'type': 'foo', 'value': 'hello world!'},
                {'value': 123},
            ]})

    def test_using_as_nested_schema_with_many(self):
        class SchemaWithMany(Schema):
            items = fields.Nested(MySchema, many=True)

        schema = SchemaWithMany()
        result = schema.load({'items': [
            {'type': 'foo', 'value': 'hello world!'},
            {'type': 'bar', 'value': 123},
        ]})
        assert {'items': [FooConfig('hello world!'), BarConfig(123)]} == result

        with self.assertRaises(ValidationError):
            schema.load({'items': [
                {'type': 'foo', 'value': 'hello world!'},
                {'value': 123},
            ]})

    def test_using_custom_type_field(self):
        class MyCustomTypeFieldSchema(MySchema):
            TYPE_FIELD = 'object_type'

        schema = MyCustomTypeFieldSchema()
        data = [FooConfig('hello'), BarConfig(111)]
        marshalled = schema.dump(data, many=True)
        assert [{'object_type': 'foo', 'value': 'hello'},
                {'object_type': 'bar', 'value': 111}] == marshalled

        unmarshalled = schema.load(marshalled, many=True)
        assert data == unmarshalled



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from hestia.tz_utils import local_now
from marshmallow import ValidationError

from polyaxon_schemas.polyflow.executable import ExecutableConfig


class TestExecutableConfigs(TestCase):
    def test_executable(self):
        config_dict = {'start_at': 'foo'}
        with self.assertRaises(ValidationError):
            ExecutableConfig.from_dict(config_dict)

        config_dict = {'execute_at': 'foo'}
        with self.assertRaises(ValidationError):
            ExecutableConfig.from_dict(config_dict)

        config_dict = {
            'timeout': 2,
            'execute_at': local_now().isoformat()
        }
        ExecutableConfig.from_dict(config_dict)

        config_dict = {
            'timeout': 2,
        }
        ExecutableConfig.from_dict(config_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import six

from marshmallow import post_dump, post_load
from marshmallow.utils import _Missing


class UnknownSchemaMixin(object):
    @post_load(pass_original=True)
    def _handle_load_unknown(self, data, original):
        """Preserve unknown keys during deserialization."""
        for key, val in original.items():
            if key not in self.fields:
                data[key] = val
        return data

    @post_dump(pass_original=True)
    def _handle_dump_unknown(self, data, original):
        """Preserve unknown keys during deserialization."""
        for key, val in original.items():
            if key not in self.fields:
                data[key] = val
        return data


missing = _Missing()


def get_value(key, obj, default=missing):
    """Helper for pulling a keyed value off various types of objects"""
    if isinstance(key, int):
        return _get_value_for_key(key, obj, default)
    return _get_value_for_keys(key.split('.'), obj, default)


def _get_value_for_keys(keys, obj, default):
    if len(keys) == 1:
        return _get_value_for_key(keys[0], obj, default)
    return _get_value_for_keys(keys[1:], _get_value_for_key(keys[0], obj, default), default)


def _get_value_for_key(key, obj, default):
    try:
        return obj[key]
    except (KeyError, AttributeError, IndexError, TypeError):
        try:
            attr = getattr(obj, key)
            if callable(attr):
                if hasattr(attr, 'get_config') or hasattr(attr, 'SCHEMA'):
                    return attr

                return attr.__name__

            return attr
        except AttributeError:
            return default


def to_camel_case(snake_str):
    split_str = snake_str.split('_')
    if len(split_str) == 1:
        try:
            return snake_str if str.isupper(snake_str[0]) else snake_str.title()
        except TypeError:
            return snake_str if six.text_type.isupper(snake_str[0]) else snake_str.title()
    return "".join(x.title() for x in split_str)


class TaskType(object):
    MASTER = 'master'
    CHIEF = 'chief'
    PS = 'ps'
    WORKER = 'worker'
    SERVER = 'server'
    SCHEDULER = 'scheduler'

    VALUES = [MASTER, PS, WORKER, SERVER, SCHEDULER]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function


class BuildBackend(object):
    NATIVE = 'native'
    KANIKO = 'kaniko'
    OTHER = 'other'

    VALUES = [NATIVE, KANIKO, OTHER]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.fields import IntOrStr


class OutputsSchema(BaseSchema):
    jobs = fields.List(IntOrStr(), allow_none=True)
    experiments = fields.List(IntOrStr(), allow_none=True)

    @staticmethod
    def schema_config():
        return OutputsConfig


class OutputsConfig(BaseConfig):
    """
    Outputs config.

    Defines the list of previous jobs/experiments outputs paths
    to make available to other experiments/jobs.

    Args:
        jobs: `list(str)`. The list of the names of jobs to make available for the current run.
        experiments: `list(str)`. The list of the names of experiments
            to make available for the current run.
    """
    IDENTIFIER = 'outputs'
    SCHEMA = OutputsSchema

    def __init__(self, jobs=None, experiments=None):
        self.jobs = jobs
        self.experiments = experiments



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, validate

from polyaxon_schemas.base import BaseConfig, BaseSchema


class Optimization(object):
    MAXIMIZE = 'maximize'
    MINIMIZE = 'minimize'

    MAXIMIZE_VALUES = [MAXIMIZE, MAXIMIZE.upper(), MAXIMIZE.capitalize()]
    MINIMIZE_VALUES = [MINIMIZE, MINIMIZE.upper(), MINIMIZE.capitalize()]

    VALUES = MAXIMIZE_VALUES + MINIMIZE_VALUES

    @classmethod
    def maximize(cls, value):
        return value in cls.MAXIMIZE_VALUES

    @classmethod
    def minimize(cls, value):
        return value in cls.MINIMIZE_VALUES


class SearchMetricSchema(BaseSchema):
    name = fields.Str()
    optimization = fields.Str(allow_none=True, validate=validate.OneOf(Optimization.VALUES))

    @staticmethod
    def schema_config():
        return SearchMetricConfig


class SearchMetricConfig(BaseConfig):
    SCHEMA = SearchMetricSchema
    IDENTIFIER = 'search_metric'

    def __init__(self,
                 name,
                 optimization=Optimization.MAXIMIZE):
        self.name = name
        self.optimization = optimization



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import os

from unittest import TestCase

from flaky import flaky

from polyaxon_schemas.ops.build_job import BuildConfig
from polyaxon_schemas.ops.environments.pods import EnvironmentConfig
from polyaxon_schemas.ops.environments.resources import K8SResourcesConfig, PodResourcesConfig
from polyaxon_schemas.ops.experiment.frameworks import ExperimentFramework
from polyaxon_schemas.ops.group.early_stopping_policies import EarlyStoppingConfig
from polyaxon_schemas.ops.group.hptuning import HPTuningConfig, SearchAlgorithms
from polyaxon_schemas.ops.group.matrix import MatrixConfig
from polyaxon_schemas.ops.logging import LoggingConfig
from polyaxon_schemas.polyaxonfile import PolyaxonFile
from polyaxon_schemas.specs.frameworks import TensorflowSpecification
from polyaxon_schemas.utils import TaskType


class TestPolyaxonfileDeprecation(TestCase):

    def test_simple_file_framework_passes(self):
        plxfile = PolyaxonFile(os.path.abspath('tests/fixtures/deprecated/simple_file_framework.yml'))
        spec = plxfile.specification
        spec.parse_data()
        assert spec.version == 1
        assert spec.logging is None
        assert spec.tags is None
        assert spec.build.dockerfile == 'Dockerfile'
        assert spec.run.cmd == 'video_prediction_train --model=DNA --num_masks=1'
        assert spec.environment is not None
        assert spec.environment.resources.gpu.to_dict() == {'requests': 1, 'limits': 1}
        assert spec.environment.outputs.to_dict() == {'jobs': [111], 'experiments': None}
        assert spec.framework is not None
        assert spec.is_experiment is True

    def test_deprecated_advanced_file_passes(self):
        plxfile = PolyaxonFile(os.path.abspath('tests/fixtures/deprecated/advanced_file.yml'))
        spec = plxfile.specification
        spec.parse_data()
        assert spec.version == 1
        assert isinstance(spec.logging, LoggingConfig)
        assert spec.is_experiment
        assert isinstance(spec.environment, EnvironmentConfig)
        assert spec.framework == ExperimentFramework.TENSORFLOW
        assert spec.config.tensorflow.n_workers == 5
        assert spec.config.tensorflow.n_ps == 10

        # check properties for returning worker configs and resources
        assert spec.config.tensorflow.worker_resources == {}
        assert spec.config.tensorflow.ps_resources == {}

        cluster, is_distributed = spec.cluster_def

        assert TensorflowSpecification.get_worker_resources(
            environment=spec.config.tensorflow,
            cluster=cluster,
            is_distributed=is_distributed
        ) == {}
        assert TensorflowSpecification.get_ps_resources(
            environment=spec.config.tensorflow,
            cluster=cluster,
            is_distributed=is_distributed
        ) == {}

        assert spec.cluster_def == ({TaskType.MASTER: 1,
                                     TaskType.WORKER: 5,
                                     TaskType.PS: 10}, True)

    def test_deprecated_notebook_job_with_node_selectors(self):
        plxfile = PolyaxonFile(os.path.abspath(
            'tests/fixtures/deprecated/notebook_with_custom_environment.yml'))
        spec = plxfile.specification
        spec.parse_data()
        assert spec.version == 1
        assert spec.is_notebook
        assert spec.is_notebook is True
        assert spec.backend is None
        assert spec.logging is None
        assert sorted(spec.tags) == sorted(['foo', 'bar'])
        assert isinstance(spec.build, BuildConfig)
        assert isinstance(spec.environment, EnvironmentConfig)
        assert spec.artifact_refs == ['outputs1']
        assert spec.data_refs == ['data1', 'data2']
        assert spec.config_map_refs == ['config_map1', 'config_map2']

        node_selector = {'polyaxon.com': 'node_for_notebook_jobs'}
        assert spec.environment.node_selector == node_selector
        assert spec.node_selector == node_selector

        resources = {
            'cpu': {'requests': 1, 'limits': 2},
            'memory': {'requests': 200, 'limits': 200},
        }
        assert spec.environment.resources.to_dict() == resources
        assert spec.resources.to_dict() == resources

        affinity = {
            'nodeAffinity': {'requiredDuringSchedulingIgnoredDuringExecution': {}}
        }
        assert spec.environment.affinity == affinity
        assert spec.affinity == affinity

        tolerations = [{'key': 'key', 'operator': 'Exists'}]

        assert spec.environment.tolerations == tolerations
        assert spec.tolerations == tolerations

    def test_deprecated_advanced_file_with_custom_configs_and_resources_passes(self):
        plxfile = PolyaxonFile(os.path.abspath(
            'tests/fixtures/deprecated/advanced_file_with_custom_configs_and_resources.yml'))
        spec = plxfile.specification
        spec.parse_data()
        assert spec.version == 1
        assert isinstance(spec.logging, LoggingConfig)
        assert spec.is_experiment
        assert isinstance(spec.environment, EnvironmentConfig)
        assert spec.framework == ExperimentFramework.TENSORFLOW
        assert spec.artifact_refs == ['outputs1']
        assert spec.data_refs == ['data1', 'data2']
        assert spec.config_map_refs == ['config_map1', 'config_map2']
        assert spec.config.tensorflow.n_workers == 5
        assert spec.config.tensorflow.n_ps == 10

        assert isinstance(spec.environment.resources, PodResourcesConfig)
        assert isinstance(spec.environment.resources.cpu, K8SResourcesConfig)
        assert spec.environment.resources.cpu.requests == 1
        assert spec.environment.resources.cpu.limits == 2

        assert spec.config.tensorflow.default_worker_node_selector == {
            'foo': True
        }

        assert spec.config.tensorflow.worker_resources == {}
        assert spec.config.tensorflow.worker_affinities == {}
        assert isinstance(spec.config.tensorflow.worker_node_selectors[3], dict)
        assert spec.config.tensorflow.worker_node_selectors[3] == {
            'foo': False
        }
        assert isinstance(spec.config.tensorflow.worker_tolerations[4], list)
        assert spec.config.tensorflow.worker_tolerations[4] == [{
            'key': 'key',
            'operator': 'Exists',
            'effect': 'NoSchedule',
        }]

        assert isinstance(spec.config.tensorflow.default_ps_resources, PodResourcesConfig)
        assert isinstance(spec.config.tensorflow.default_ps_resources.cpu, K8SResourcesConfig)
        assert spec.config.tensorflow.default_ps_resources.cpu.requests == 2
        assert spec.config.tensorflow.default_ps_resources.cpu.limits == 4

        assert spec.config.tensorflow.ps_node_selectors == {}
        assert isinstance(spec.config.tensorflow.ps_tolerations[7], list)
        assert spec.config.tensorflow.ps_tolerations[7] == [{
            'operator': 'Exists'
        }]
        assert isinstance(spec.config.tensorflow.ps_affinities[7], dict)
        assert isinstance(spec.config.tensorflow.ps_resources[9], PodResourcesConfig)
        assert isinstance(spec.config.tensorflow.ps_resources[9].memory, K8SResourcesConfig)
        assert spec.config.tensorflow.ps_resources[9].memory.requests == 512
        assert spec.config.tensorflow.ps_resources[9].memory.limits == 1024

        # check that properties for return list of configs and resources is working
        cluster, is_distributed = spec.cluster_def
        worker_node_selectors = TensorflowSpecification.get_worker_node_selectors(
            environment=spec.config.tensorflow,
            cluster=cluster,
            is_distributed=is_distributed
        )
        assert len(worker_node_selectors) == spec.config.tensorflow.n_workers
        assert set([i['foo'] for i in worker_node_selectors.values()]) == {
            spec.config.tensorflow.default_worker_node_selector['foo'],
            spec.config.tensorflow.worker_node_selectors[3]['foo']}

        assert TensorflowSpecification.get_worker_resources(
            environment=spec.config.tensorflow,
            cluster=cluster,
            is_distributed=is_distributed
        ) == {}
        ps_resources = TensorflowSpecification.get_ps_resources(
            environment=spec.config.tensorflow,
            cluster=cluster,
            is_distributed=is_distributed
        )
        assert len(ps_resources) == spec.config.tensorflow.n_ps
        assert set(ps_resources.values()) == {
            spec.config.tensorflow.default_ps_resources,
            spec.config.tensorflow.ps_resources[9]}

        # Check total resources
        assert spec.total_resources == {
            'cpu': {'requests': 1 + 2 * 9, 'limits': 2 + 4 * 9},
            'memory': {'requests': 512, 'limits': 1024},
        }

        assert spec.cluster_def == ({TaskType.MASTER: 1,
                                     TaskType.WORKER: 5,
                                     TaskType.PS: 10}, True)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from marshmallow import ValidationError

from polyaxon_schemas.ops import params as ops_params
from polyaxon_schemas.ops.io import IOTypes
from polyaxon_schemas.ops.operation import BaseOpConfig


class TestOperationsConfigs(TestCase):
    def test_passing_params_declarations_raises(self):
        config_dict = {
            'params': {'foo': 'bar'},
            'declarations': {'foo': 'bar'},
        }

        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

    def test_passing_declarations_sets_params(self):
        config_dict = {
            'declarations': {'foo': 'bar'},
        }

        op = BaseOpConfig.from_dict(config_dict)
        assert op.params == config_dict['declarations']

    def test_passing_params_passes(self):
        config_dict = {
            'params': {'foo': 'bar'},
        }

        op = BaseOpConfig.from_dict(config_dict)
        assert op.params == config_dict['params']

    def test_param_validation_with_inputs(self):
        config_dict = {
            'params': {
                'param1': 'text',
                'param2': 12,
                'param3': 13.3,
                'param4': False,
                'param5': {'foo': 'bar'},
                'param6': [1, 3, 45, 5],
                'param7': 'gs://bucket/path/to/blob/',
                'param8': 's3://test/this/is/bad/key.txt',
                'param9': 'wasbs://container@user.blob.core.windows.net/',
                'param10': '/foo/bar'
            },
            'inputs': [
                {'name': 'param1', 'type': IOTypes.STR},
                {'name': 'param2', 'type': IOTypes.INT},
                {'name': 'param3', 'type': IOTypes.FLOAT},
                {'name': 'param4', 'type': IOTypes.BOOL},
                {'name': 'param5', 'type': IOTypes.DICT},
                {'name': 'param6', 'type': IOTypes.LIST},
                {'name': 'param7', 'type': IOTypes.GCS_PATH},
                {'name': 'param8', 'type': IOTypes.S3_PATH},
                {'name': 'param9', 'type': IOTypes.AZURE_PATH},
                {'name': 'param10', 'type': IOTypes.PATH},
            ]
        }
        op = BaseOpConfig.from_dict(config_dict)
        assert op.params == config_dict['params']

    def test_param_validation_with_outputs(self):
        config_dict = {
            'params': {
                'param1': 'text',
                'param2': 12,
                'param3': 13.3,
                'param4': False,
                'param5': {'foo': 'bar'},
                'param6': [1, 3, 45, 5],
                'param7': 'gs://bucket/path/to/blob/',
                'param8': 's3://test/this/is/bad/key.txt',
                'param9': 'wasbs://container@user.blob.core.windows.net/',
                'param10': '/foo/bar',
                'param11': 124.4,
                'param12': {'foo': 124.4},
                'param13': {'foo': 'bar'},
                'param14': {'foo': ['foo', 124.4]},
            },
            'outputs': [
                {'name': 'param1', 'type': IOTypes.STR},
                {'name': 'param2', 'type': IOTypes.INT},
                {'name': 'param3', 'type': IOTypes.FLOAT},
                {'name': 'param4', 'type': IOTypes.BOOL},
                {'name': 'param5', 'type': IOTypes.DICT},
                {'name': 'param6', 'type': IOTypes.LIST},
                {'name': 'param7', 'type': IOTypes.GCS_PATH},
                {'name': 'param8', 'type': IOTypes.S3_PATH},
                {'name': 'param9', 'type': IOTypes.AZURE_PATH},
                {'name': 'param10', 'type': IOTypes.PATH},
                {'name': 'param11', 'type': IOTypes.METRIC},
                {'name': 'param12', 'type': IOTypes.METADATA},
                {'name': 'param13', 'type': IOTypes.METADATA},
                {'name': 'param14', 'type': IOTypes.METADATA},
            ]
        }
        op = BaseOpConfig.from_dict(config_dict)
        assert op.params == config_dict['params']

    def test_required_input_no_param_only_validated_on_run(self):
        # Inputs
        config_dict = {
            'params': {
                'param1': 'text',
            },
            'inputs': [
                {'name': 'param1', 'type': IOTypes.STR},
                {'name': 'param10', 'type': IOTypes.PATH},
            ]
        }
        config = BaseOpConfig.from_dict(config_dict)
        with self.assertRaises(ValidationError):
            ops_params.validate_params(params=config.params,
                                       inputs=config.inputs,
                                       outputs=config.outputs,
                                       is_template=False,
                                       is_run=True)

        # Outputs
        config_dict = {
            'params': {
                'param1': 'text',
            },
            'outputs': [
                {'name': 'param1', 'type': IOTypes.STR},
                {'name': 'param10', 'type': IOTypes.PATH},
            ]
        }
        config = BaseOpConfig.from_dict(config_dict)

        ops_params.validate_params(params=config.params,
                                   inputs=config.inputs,
                                   outputs=config.outputs,
                                   is_template=False,
                                   is_run=True)

        # IO
        config_dict = {
            'params': {
                'param1': 'text',
            },
            'inputs': [
                {'name': 'param1', 'type': IOTypes.STR},
            ],
            'outputs': [
                {'name': 'param10', 'type': IOTypes.PATH},
            ]
        }
        config = BaseOpConfig.from_dict(config_dict)
        ops_params.validate_params(params=config.params,
                                   inputs=config.inputs,
                                   outputs=config.outputs,
                                   is_template=False,
                                   is_run=True)

    def test_incomplete_params(self):
        config_dict = {
            'params': {
                'param1': 1,
            },
            'inputs': [
                {'name': 'param1', 'type': IOTypes.INT},
                {'name': 'param2', 'type': IOTypes.INT},
            ]
        }
        config = BaseOpConfig.from_dict(config_dict)
        with self.assertRaises(ValidationError):
            ops_params.validate_params(params=config.params,
                                       inputs=config.inputs,
                                       outputs=config.outputs,
                                       is_template=False,
                                       is_run=True)

        config_dict = {
            'params': {
                'param1': 1,
            },
            'outputs': [
                {'name': 'param1', 'type': IOTypes.INT, 'default': 12, 'is_optional': True},
                {'name': 'param2', 'type': IOTypes.INT},
            ]
        }
        config = BaseOpConfig.from_dict(config_dict)
        ops_params.validate_params(params=config.params,
                                   inputs=config.inputs,
                                   outputs=config.outputs,
                                   is_template=False,
                                   is_run=True)

    def test_extra_params(self):
        # inputs
        config_dict = {
            'params': {
                'param1': 1,
                'param2': 2,
            },
            'inputs': [
                {'name': 'param1', 'type': IOTypes.INT},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        # outputs
        config_dict = {
            'params': {
                'param1': 1,
                'param2': 2,
            },
            'outputs': [
                {'name': 'param1', 'type': IOTypes.INT},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

    def test_param_validation_with_mismatched_inputs(self):
        config_dict = {
            'params': {
                'param1': 'text',
            },
            'inputs': [
                {'name': 'param1', 'type': IOTypes.INT},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        config_dict = {
            'params': {
                'param2': 12,
            },
            'inputs': [
                {'name': 'param2', 'type': IOTypes.STR},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        config_dict = {
            'params': {
                'param3': 13.3,
            },
            'inputs': [
                {'name': 'param3', 'type': IOTypes.INT},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        config_dict = {
            'params': {
                'param4': False,
            },
            'inputs': [
                {'name': 'param4', 'type': IOTypes.STR},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        config_dict = {
            'params': {
                'param5': {'foo': 'bar'},
            },
            'inputs': [
                {'name': 'param5', 'type': IOTypes.STR},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        config_dict = {
            'params': {
                'param6': [1, 3, 45, 5],
            },
            'inputs': [
                {'name': 'param6', 'type': IOTypes.STR},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        config_dict = {
            'params': {
                'param7': 'gs://bucket/path/to/blob/',
            },
            'inputs': [
                {'name': 'param7', 'type': IOTypes.INT},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        config_dict = {
            'params': {
                'param8': 's3://test/this/is/bad/key.txt',
            },
            'inputs': [
                {'name': 'param8', 'type': IOTypes.AZURE_PATH},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        config_dict = {
            'params': {
                'param9': 'wasbs://container@user.blob.core.windows.net/',
            },
            'inputs': [
                {'name': 'param9', 'type': IOTypes.FLOAT},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

    def test_param_validation_with_mismatched_outputs(self):
        config_dict = {
            'params': {
                'param1': 'text',
            },
            'outputs': [
                {'name': 'param1', 'type': IOTypes.INT},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        config_dict = {
            'params': {
                'param2': 12,
            },
            'outputs': [
                {'name': 'param2', 'type': IOTypes.STR},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        config_dict = {
            'params': {
                'param3': 13.3,
            },
            'outputs': [
                {'name': 'param3', 'type': IOTypes.INT},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        config_dict = {
            'params': {
                'param4': False,
            },
            'outputs': [
                {'name': 'param4', 'type': IOTypes.STR},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        config_dict = {
            'params': {
                'param5': {'foo': 'bar'},
            },
            'outputs': [
                {'name': 'param5', 'type': IOTypes.STR},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        config_dict = {
            'params': {
                'param6': [1, 3, 45, 5],
            },
            'outputs': [
                {'name': 'param6', 'type': IOTypes.STR},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        config_dict = {
            'params': {
                'param7': 'gs://bucket/path/to/blob/',
            },
            'outputs': [
                {'name': 'param7', 'type': IOTypes.INT},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        config_dict = {
            'params': {
                'param8': 's3://test/this/is/bad/key.txt',
            },
            'outputs': [
                {'name': 'param8', 'type': IOTypes.AZURE_PATH},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        config_dict = {
            'params': {
                'param9': 'wasbs://container@user.blob.core.windows.net/',
            },
            'outputs': [
                {'name': 'param9', 'type': IOTypes.FLOAT},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        config_dict = {
            'params': {
                'param12': 'wasbs://container@user.blob.core.windows.net/',
            },
            'outputs': [
                {'name': 'param12', 'type': IOTypes.METADATA},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

        config_dict = {
            'params': {
                'param9': 'wasbs://container@user.blob.core.windows.net/',
            },
            'outputs': [
                {'name': 'param9', 'type': IOTypes.METRIC},
            ]
        }
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)

    def test_experiment_and_job_refs_params(self):
        config_dict = {
            'params': {
                'param1': '{{ experiments.1.outputs.foo }}',
                'param2': '{{ jobs.1.outputs.foo }}',
                'param9': 'wasbs://container@user.blob.core.windows.net/',
                'param11': '{{ experiments.1.outputs.accuracy }}',
            },
            'inputs': [
                {'name': 'param1', 'type': IOTypes.INT},
                {'name': 'param2', 'type': IOTypes.FLOAT},
                {'name': 'param9', 'type': IOTypes.AZURE_PATH},
                {'name': 'param11', 'type': IOTypes.METRIC},
            ],
        }
        op = BaseOpConfig.from_dict(config_dict)
        assert op.params == config_dict['params']

    def test_op_refs_params(self):
        config_dict = {
            'params': {
                'param1': '{{ ops.A.outputs.foo }}',
                'param9': 13.1,
            },
            'inputs': [
                {'name': 'param1', 'type': IOTypes.INT},
                {'name': 'param9', 'type': IOTypes.FLOAT},
            ]
        }
        # Validation outside the context of a pipeline
        with self.assertRaises(ValidationError):
            BaseOpConfig.from_dict(config_dict)



EXPERIMENT = 'experiment'
GROUP = 'group'
JOB = 'job'
NOTEBOOK = 'notebook'
TENSORBOARD = 'tensorboard'
BUILD = 'build'
PIPELINE = 'pipeline'
KINDS = {EXPERIMENT, GROUP, JOB, NOTEBOOK, TENSORBOARD, BUILD, PIPELINE}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseMultiSchema, BaseSchema
from polyaxon_schemas.fields import ObjectOrListObject


class GlobalStepWaiterHookSchema(BaseSchema):
    wait_until_step = fields.Int()

    @staticmethod
    def schema_config():
        return GlobalStepWaiterHookConfig


class GlobalStepWaiterHookConfig(BaseConfig):
    IDENTIFIER = 'GlobalStepWaiterHook'
    SCHEMA = GlobalStepWaiterHookSchema

    def __init__(self, wait_until_step):
        self.wait_until_step = wait_until_step


class FinalOpsHookSchema(BaseSchema):
    final_ops = ObjectOrListObject(fields.Str)

    @staticmethod
    def schema_config():
        return FinalOpsHookConfig


class FinalOpsHookConfig(BaseConfig):
    IDENTIFIER = 'FinalOpsHook'
    SCHEMA = FinalOpsHookSchema

    def __init__(self, final_ops):
        self.final_ops = final_ops


class StepLoggingTensorHookSchema(BaseSchema):
    tensors = ObjectOrListObject(fields.Str)
    every_n_iter = fields.Int(allow_none=True)
    every_n_secs = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return StepLoggingTensorHookConfig


class StepLoggingTensorHookConfig(BaseConfig):
    IDENTIFIER = 'StepLoggingTensorHook'
    SCHEMA = StepLoggingTensorHookSchema

    def __init__(self, tensors, every_n_iter=None, every_n_secs=None):
        self.tensors = tensors
        self.every_n_iter = every_n_iter
        self.every_n_secs = every_n_secs


class EpisodeLoggingTensorHookSchema(BaseSchema):
    tensors = ObjectOrListObject(fields.Str)
    every_n_episodes = fields.Int()

    @staticmethod
    def schema_config():
        return EpisodeLoggingTensorHookConfig


class EpisodeLoggingTensorHookConfig(BaseConfig):
    IDENTIFIER = 'EpisodeLoggingTensorHook'
    SCHEMA = EpisodeLoggingTensorHookSchema

    def __init__(self, tensors, every_n_episodes):
        self.tensors = tensors
        self.every_n_episodes = every_n_episodes


class HookSchema(BaseMultiSchema):
    __multi_schema_name__ = 'loss'
    __configs__ = {
        GlobalStepWaiterHookConfig.IDENTIFIER: GlobalStepWaiterHookConfig,
        FinalOpsHookConfig.IDENTIFIER: FinalOpsHookConfig,
        StepLoggingTensorHookConfig.IDENTIFIER: StepLoggingTensorHookConfig,
        EpisodeLoggingTensorHookConfig.IDENTIFIER: EpisodeLoggingTensorHookConfig,
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from collections import Mapping

from marshmallow import fields


class DictOrStr(fields.Str):
    default_error_messages = {
        'invalid': 'Not a valid string or dict.',
        'invalid_utf8': 'Not a valid utf-8 string.'
    }

    def _serialize(self, value, attr, obj, **kwargs):
        if isinstance(value, Mapping):
            return value

        return super(DictOrStr, self)._serialize(value=value, attr=attr, obj=obj)

    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, Mapping):
            return value

        return super(DictOrStr, self)._deserialize(value=value, attr=attr, data=data)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_dict

from polyaxon_schemas.ops.environments.pods import EnvironmentConfig
from polyaxon_schemas.ops.environments.resources import K8SResourcesConfig, PodResourcesConfig


class TestEnvironmentsConfigs(TestCase):

    def test_environment_config(self):
        config_dict = {
            'resources': PodResourcesConfig(cpu=K8SResourcesConfig(0.5, 1)).to_dict()
        }
        config = EnvironmentConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        # Add node selectors
        config_dict['node_selector'] = {
            'polyaxon.com': 'master',
        }

        config = EnvironmentConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        # Add persistence
        config_dict['data_refs'] = ['data1', 'data2']
        config_dict['artifact_refs'] = ['outputs1']
        config = EnvironmentConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        # TODO: Add outputs
        # config_dict['outputs'] = {
        #     'jobs': ['data1.dfs', 34, 'data2'],
        #     'experiments': [1, 'outputs1', 2, 3],
        # }
        # config = EnvironmentConfig.from_dict(config_dict)
        # assert_equal_dict(config_dict, config.to_dict())

        # Add secrets
        config_dict['secret_refs'] = ['secret1', 'secret2']
        config = EnvironmentConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        # Add config_maps
        config_dict['config_map_refs'] = ['config_map1', 'config_map2']
        config = EnvironmentConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        # Add service_account
        config_dict['service_account'] = 'service_account'
        config = EnvironmentConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        # Add image_pull_secrets
        config_dict['image_pull_secrets'] = ['pull_secret1', 'pull_secret2']
        config = EnvironmentConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        # Add max_restarts
        config_dict['max_restarts'] = 4
        config = EnvironmentConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        # Add data_refs
        config_dict['data_refs'] = ['data1', 'data2']
        config = EnvironmentConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        # Add artifact_refs
        config_dict['artifact_refs'] = ['artifact1', 'artifact2']
        config = EnvironmentConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from marshmallow import ValidationError

from polyaxon_schemas.polyflow.ops import OpConfig


class TestOpConfigs(TestCase):
    def test_op_raises_for_template_action_event(self):
        config_dict = {'template': {'action': 'foo', 'event': 'bar'}}
        with self.assertRaises(ValidationError):
            OpConfig.from_dict(config_dict)

        config_dict = {'template': {'path': 'bar', 'action': 'foo', }}
        with self.assertRaises(ValidationError):
            OpConfig.from_dict(config_dict)

        config_dict = {'template': {'url': 'bar', 'action': 'foo'}}
        with self.assertRaises(ValidationError):
            OpConfig.from_dict(config_dict)

    def test_op_passes_for_template_action_event(self):
        config_dict = {'template': {'event': 'bar'}}
        config = OpConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

        config_dict = {'template': {'action': 'foo'}}
        config = OpConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

        config_dict = {'template': {'path': 'bar'}}
        config = OpConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

        config_dict = {'template': {'url': 'bar'}}
        config = OpConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

        config_dict = {'template': {'name': 'bar'}}
        config = OpConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_validation(self):
        config_dict = {'template': {'action': 'foo'}, 'concurrency': 'foo'}
        with self.assertRaises(ValidationError):
            OpConfig.from_dict(config_dict)

        config_dict = {'template': {'action': 'foo'}, 'dependencies': 'foo'}
        with self.assertRaises(ValidationError):
            OpConfig.from_dict(config_dict)

        config_dict = {
            'template': {'action': 'foo'},
            'concurrency': 2,
            'dependencies': ['foo', 'bar'],
            'params': {'param1': 'foo', 'param2': 'bar'},
            'trigger': 'all_succeeded',
            'max_retries': 4,
        }
        config = OpConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

        config_dict = {
            'template': {'event': 'foo'},
            'concurrency': 2,
        }
        config = OpConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

        config_dict = {
            'template': {'name': 'foo'},
            'concurrency': 2,
            'dependencies': [{'name': 'foo'}, {'name': 'bar'}],  # Wrong
            'params': {'param1': 'foo', 'param2': 'bar'},
            'trigger': 'all_succeeded',
            'max_retries': 4,
        }
        with self.assertRaises(ValidationError):
            OpConfig.from_dict(config_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.ml.layers.base import BaseLayerConfig, BaseLayerSchema


class WrapperSchema(BaseLayerSchema):
    layer = fields.Nested('LayerSchema')

    @staticmethod
    def schema_config():
        return WrapperConfig


class WrapperConfig(BaseLayerConfig):
    """Abstract wrapper base class.

    Wrappers take another layer and augment it in various ways.
    Do not use this class as a layer, it is only an abstract base class.
    Two usable wrappers are the `TimeDistributed` and `Bidirectional` wrappers.

    Args:
        layer: The layer to be wrapped.
    """
    IDENTIFIER = 'Wrapper'
    SCHEMA = WrapperSchema

    def __init__(self, layer, **kwargs):
        super(WrapperConfig, self).__init__(**kwargs)
        self.layer = layer


class TimeDistributedSchema(WrapperSchema):
    @staticmethod
    def schema_config():
        return TimeDistributedConfig


class TimeDistributedConfig(WrapperConfig):
    """This wrapper allows to apply a layer to every temporal slice of an input.

    The input should be at least 3D, and the dimension of index one
    will be considered to be the temporal dimension.

    Consider a batch of 32 samples,
    where each sample is a sequence of 10 vectors of 16 dimensions.
    The batch input shape of the layer is then `(32, 10, 16)`,
    and the `input_shape`, not including the samples dimension, is `(10, 16)`.

    You can then use `TimeDistributed` to apply a `Dense` layer
    to each of the 10 timesteps, independently:

    ```python
    # as the first layer in a model
    x = TimeDistributed(Dense(8))(x)
    # now x.output_shape == (None, 10, 8)
    ```

    The output will then have shape `(32, 10, 8)`.

    In subsequent layers, there is no need for the `input_shape`:

    ```python
    x = TimeDistributed(Dense(32))(x)
    # now x.output_shape == (None, 10, 32)
    ```

    The output will then have shape `(32, 10, 32)`.

    `TimeDistributed` can be used with arbitrary layers, not just `Dense`,
    for instance with a `Conv2D` layer:

    ```python
    x = TimeDistributed(Conv2D(64, (3, 3)))(x)
    ```

    Args:
        layer: a layer instance.

    Polyaxonfile usage:

    ```yaml
    TimeDistributed:
      layer:
        Dense:
          units: 2
    ```
    """
    IDENTIFIER = 'TimeDistributed'
    SCHEMA = TimeDistributedSchema


class BidirectionalSchema(WrapperSchema):

    @staticmethod
    def schema_config():
        return BidirectionalConfig


class BidirectionalConfig(WrapperConfig):
    """Bidirectional wrapper for RNNs.

    Args:
        layer: `Recurrent` instance.
        merge_mode: Mode by which outputs of the
            forward and backward RNNs will be combined.
            One of {'sum', 'mul', 'concat', 'ave', None}.
            If None, the outputs will not be combined,
            they will be returned as a list.

    Raises:
        ValueError: In case of invalid `merge_mode` argument.

    Example:

    ```python
    x = Bidirectional(plx.layers.LSTM(units=128, dropout=0.2, recurrent_dropout=0.2))(x)
    ```

    Polyaxonfile usage:

    ```yaml
    Bidirectional:
      layer:
        LSTM:
          units: 128
          dropout: 0.2
          recurrent_dropout: 0.2
    ```
    """
    IDENTIFIER = 'Bidirectional'
    SCHEMA = BidirectionalSchema



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import re
import six

from collections import namedtuple

from marshmallow import ValidationError

REGEX = re.compile(r'{{\s*([^\s]*)\s*}}')

OUTPUTS = 'outputs'
OPS = 'ops'
EXPERIMENTS = 'experiments'
JOBS = 'jobs'
ENTITIES = {JOBS, EXPERIMENTS, OPS}


class ParamSpec(namedtuple("ParamSpec", "name iotype value entity entity_ref is_flag")):
    pass


def get_param_display_value(param, value):
    if param.is_flag:
        return '--{}'.format(param.name) if value else ''
    return value


def get_param(name, value, iotype, is_flag):
    """
    Checks if the value is param ref and validates it.

    returns: ParamSpec or None
    raises: ValidationError
    """
    if not isinstance(value, six.string_types):
        return ParamSpec(name=name,
                         iotype=iotype,
                         value=value,
                         entity=None,
                         entity_ref=None,
                         is_flag=is_flag)

    param = REGEX.search(value)
    if not param:
        return ParamSpec(name=name,
                         iotype=iotype,
                         value=value,
                         entity=None,
                         entity_ref=None,
                         is_flag=is_flag)

    param = param.group(1)
    param_parts = param.split('.')
    if len(param_parts) < 3 or len(param_parts) > 4:
        raise ValidationError('Could not parse value `{}` for param `{}`.'.format(value, name))
    if param_parts[2] != OUTPUTS:
        raise ValidationError('Param `{}` value `{}` is not valid, '
                              'it should follow a format '
                              '`entity.entity-id.outputs.name`.'.format(name, value))
    if param_parts[0] not in ENTITIES:
        raise ValidationError('Could not parse value `{}` for param `{}`.'.format(value, name))
    if param_parts[0] in {EXPERIMENTS, JOBS}:
        try:
            int(param_parts[1])
        except (KeyError, ValueError):
            raise ValidationError('Param value `{}` is not valid, '
                                  'it should provide an id.'.format(value))

    return ParamSpec(name=name,
                     iotype=iotype,
                     value=param,
                     entity=param_parts[0],
                     entity_ref=param_parts[1],
                     is_flag=is_flag)


def validate_param(param, context):
    """
    Given a param reference to an operation, we check that the operation exists in the context,
    and that the types
    """
    if param.entity != OPS:
        return

    context = context or {}

    if param.value not in context:
        raise ValidationError(
            'Param `{}` has a ref value `{}`, '
            'but op with name `{}` has no such output, '
            'please check that your pipeline defines the correct template.'.format(
                param.name, param.value, param.entity_ref
            ))

    if param.iotype != context[param.value].iotype:
        raise ValidationError(
            'Param `{}` has a an input type `{}` '
            'and it does not correspond to the output type of ref `{}.'.format(
                param.name, param.iotype, param.entity_ref
            ))


def validate_params(params, inputs, outputs, context=None, is_template=True, is_run=True):
    """
    Validates Params given inputs, and an optional context.

    Params can be:
     * plain values: we check them against the inputs types
     * job/experiment references: We postpone to server side validation.
     * ops reference: in that case a context must be provided to validate that the reference exists.
        and types are correct.
    """
    if is_run and not inputs and not outputs:
        return []

    params = params or {}
    inputs = inputs or []
    outputs = outputs or []

    processed_params = []
    validated_params = []

    for inp in inputs:
        if inp.name in params:
            param_value = params[inp.name]
            param = get_param(name=inp.name,
                              value=param_value,
                              iotype=inp.iotype,
                              is_flag=inp.is_flag)
            if param.entity_ref:
                validate_param(param, context)
            else:  # Plain value
                inp.validate_value(param_value)
            validated_params.append(param)
            processed_params.append(inp.name)
        elif not inp.is_optional and not is_template:
            raise ValidationError('Input {} is required, no param was passed.'.format(inp.name))
        else:
            validated_params.append(ParamSpec(name=inp.name,
                                              value=inp.default,
                                              iotype=inp.iotype,
                                              entity=None,
                                              entity_ref=None,
                                              is_flag=inp.is_flag))

    for out in outputs:
        if out.name in params:
            param_value = params[out.name]
            param = get_param(name=out.name,
                              value=param_value,
                              iotype=out.iotype,
                              is_flag=out.is_flag)
            validated_params.append(param)
            if param.entity_ref:
                validate_param(param, None)
            else:  # Plain value
                out.validate_value(param_value)
            validated_params.append(param)
            processed_params.append(out.name)
        # No validation for outputs we assume that the op might populate a context or send a metric
        else:
            validated_params.append(ParamSpec(name=out.name,
                                              value=out.default,
                                              iotype=out.iotype,
                                              entity=None,
                                              entity_ref=None,
                                              is_flag=out.is_flag))
    extra_params = set(six.iterkeys(params)) - set(processed_params)
    if extra_params:
        raise ValidationError('Received unexpected params `{}`.'.format(extra_params))

    return validated_params
